/*
 *  Copyright (C) 2007-08 Luca Deri <deri@ntop.org>
 *  Copyright 2007 Tilera Corporation. All Rights Reserved.
 *
 *  			 http://www.ntop.org/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* *********************************************************** */

static int create_thread(ilibGroup *spawned_procs, 
			 ThreadFunctionPointer start_routine, void *arg) {
  ThreadInit    thread;
  ilibProcParam params;

  thread.shared_vars = readWriteGlobals;
  thread.shared_read_only = readOnlyGlobals;
  memcpy(&thread.shared_read_only, &readOnlyGlobals, sizeof(ReadOnlyGlobals));

  thread.thread_func = start_routine;
  thread.argument = arg;

  params.init_block = &thread;
  params.init_size = sizeof(ThreadInit);
  params.num_procs = 1; /* Create one thread */
  params.binary_name = NULL;  /* Create a copy of myself */
  params.argv = NULL;
  params.tiles.x = 0;
  params.tiles.y = 0;
  params.tiles.width = 0;
  params.tiles.height = 0;

  if(ilib_proc_spawn(1, &params, spawned_procs) < 0) {
    ilib_die("Failed to spawn process.");
    return(-1);
  } else
    return(0);
}

/* *********************************************************** */

int _pthread_create(pthread_t *thread,
		   const pthread_attr_t *restrict attr,
		   void *(*start_routine)(void*), void *arg) {
  return(create_thread(thread, start_routine, arg));
}

/* ****************************************************** */

int createCondvar(ConditionalVariable *condvarId) {
  int rc;

  rc = ilib_mutex_init(&condvarId->mutex);
  ilib_mutex_lock(&condvarId->mutex);
  condvarId->predicate = 0;
  rc = ilib_mutex_unlock(&condvarId->mutex);

  return(rc);
}

/* ************************************ */

void deleteCondvar(ConditionalVariable *condvarId) {
  ilib_mutex_destroy(&condvarId->mutex);
}

/* ****************************************************** */

int waitCondvar(ConditionalVariable *condvarId) {
  int debug = 0;
  struct timespec ts;

  do {
    // First wait for it to become true before trying to lock it
    //
    if(debug) printf("= 1 => waitCondvar() [%p][%d][predicate=%d]\n", 
		     condvarId, 0, condvarId->predicate);

    ts.tv_sec = 0;
    ts.tv_nsec = 5000000; /* 0.5 sec */
    
    while(condvarId->predicate <= 0) {
      if(debug) printf("= LOOP => waitCondvar() [%d][predicate=%d]\n", 1, condvarId->predicate);
      //sleep(1);
      nanosleep (&ts, NULL);
      ; // loop
    }

    // Then confirm that it is true while locked
    //

    if(debug) printf("= 2 => waitCondvar() [%d][predicate=%d]\n", 1, condvarId->predicate);

    ilib_mutex_lock(&condvarId->mutex);
    if(debug) printf("= 3 => waitCondvar() [%d][predicate=%d]\n", 2, condvarId->predicate);
    if (condvarId->predicate > 0)
      break;
    if(debug) printf("= 4 => waitCondvar() [%d][predicate=%d]\n", 3, condvarId->predicate);
    ilib_mutex_unlock(&condvarId->mutex);

  } while(1);

  condvarId->predicate--;

  int rc = ilib_mutex_unlock(&condvarId->mutex);

  if(debug) printf("==> waitCondvar() returned\n");

  return rc;
}

/* ************************************ */

int signalCondvar(ConditionalVariable *condvarId, int broadcast) {
  int rc, debug = 0;

  rc = ilib_mutex_lock(&condvarId->mutex);

  condvarId->predicate++;

  if(debug) printf("==> signalCondvar(%p, predicate=%d)\n", condvarId, condvarId->predicate);
  rc = ilib_mutex_unlock(&condvarId->mutex);

  ////  if(broadcast)
  ////    rc = pthread_cond_broadcast(&condvarId->condvar);
  ////  else
  ////    rc = pthread_cond_signal(&condvarId->condvar);

  return rc;
}

